var assert = chai.assert;

function verifyAVLProof(p) {
  return true;
}

// see https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/DataView
// https://developer.mozilla.org/en-US/docs/Web/JavaScript/Typed_arrays
function intToByte(x) {
  var bigInt;
  if (x > Math.pow(2,63)) {
    bigInt = new BigUint64Array(1); 
  }
  else {
    bigInt = new BigInt64Array(1);
  }
  
  bigInt[0] = BigInt(x);
  var byteResult = new Uint8Array(bigInt.buffer).reverse();
  return byteResult;
}

function floatToByte(x) {
  var floatArr = new Float64Array(1);  
  floatArr[0] = x;
  var byteResult = new Uint8Array(floatArr.buffer).reverse();
  return byteResult;
}

function byteToFloat(x) {
  var bytes = new Uint8Array(x.reverse());
  var floatnum = new Float64Array(bytes.buffer)[0];
  
  return floatnum;
}

function byteToInt(x) {
  var bytes = new Uint8Array(x.reverse());
  var bigint = new BigInt64Array(bytes.buffer)[0];
  
  return bigint;
}

function byteToUint(x) {
  var bytes = new Uint8Array(x.reverse());
  var bigint = new BigUint64Array(bytes.buffer)[0];
  
  return bigint;
}

describe('Standalone Proofs', function() {

  it('floatToByte', function(done) {
    let f = 3.1415926535;
    let floatbytes = floatToByte(f);
    // console.log(floatbytes);
    let f2 = byteToFloat(floatbytes);
    // console.log(f2);
    assert( f == f2 );

    let g = 7139627614.31415926535;
    let gloatbytes = floatToByte(g);
    // console.log(gloatbytes);
    let g2 = byteToFloat(gloatbytes);
    // console.log(g2);
    assert( g == g2 );
    done();
  });

    it('intToByte', function(done) {
      let i = -7139627614n;
      let ibytes = intToByte(i);
      // console.log(ibytes);
      let i2 = byteToInt(ibytes);
      // console.log(i2);
      assert( i == i2 );

      let j = 18446744073709551615n; // 2^64-1
      let jbytes = intToByte(j);
      // console.log(jbytes);
      let j2 = byteToUint(jbytes);
      // console.log(j2);
      assert( j == j2 );
      done();
    });


  it('VerifyAVLProof (int)', function(done) {
    let proofObject = {"lp":[{"k":"B0D180265C3B4540","h":3,"s":5,"sbt":97,"l":"35F6D2A33128422565DE01BE919F3AB83416A83CEEF9F4D7C31949D090738F11","r":""},{"k":"E020B7BDB8C24940","h":2,"s":3,"sbt":68,"l":"","r":"D41DC24421F3B396C216FAA211E4771494924B7D58C5217A0589CBB3653DAE62"},{"k":"BD78577A69084340","h":1,"s":2,"sbt":36,"l":"2FA705480F978ABA68571EE0E876AEF28FF2BF4D11A2C0DBFD450B4226A2608A","r":""}],"in":null,"lfs":[{"k":"BD78577A69084340","v":"A78F21B7276D3540","h":0,"s":1,"sb":13}]};

    var verified = verifyAVLProof(proofObject, 1);
          assert(verified);
          done();
  });
});
