'use strict';
const crypto = window.crypto;

const BucketCloudstore = 0
const BucketNoSQL = 1
const BucketSQL = 2

const DEFAULT_DEVELOPER_TRUST_LEVEL = 1 //0 = regular web behavior no proof checking | 1 = check ALL proofs | In between check that % of time
const DEFAULT_USER_TRUST_LEVEL = 1 //0 = regular web behavior no proof checking | 1 = check ALL proofs | In between check that % of time

class WEXTENSION {

  constructor(callback) {
    this.provider = "https://cloud.wolk.com";
    this.developerTrustLevel = DEFAULT_DEVELOPER_TRUST_LEVEL;
    this.userTrustLevel = DEFAULT_USER_TRUST_LEVEL;
    this.wolkTrustLevel = 0;
    this.setWolkTrustLevel();
    this.loadAccounts(callback);
  }

  getOwner() {
    return window.localStorage.getItem("activeAccount");
  }

  getProvider() {
    return this.provider;
  }

  setProvider(provider) {
    this.provider = provider
  }

  setDeveloperTrustLevel( level ) {
    this.developerTrustLevel = level;
    this.setWolkTrustLevel();
  }

  setUserTrustLevel( level ) {
    this.userTrustLevel = level;
    this.setWolkTrustLevel();
  }

  setWolkTrustLevel() {
    if( this.userTrustLevel > this.developerTrustLevel ) {
      this.wolkTrustLevel  = this.userTrustLevel
    } else {
      this.wolkTrustLevel  = this.developerTrustLevel
    }
  }

  determineProofValue() {
    let randProof = Math.random();
    if ( randProof <= this.wolkTrustLevel ) {
      return 1
    } else {
      return 0
    }
  }

  createAccount(name, options, callback) {
    let that = this;
    console.log("createAccount", name);
    // generate + export RSA Key
    return crypto.subtle.generateKey({
      name: "RSASSA-PKCS1-v1_5",
      modulusLength: 2048,
      publicExponent: new Uint8Array([0x01, 0x00, 0x01]),
      hash: {
        name: "SHA-256"
      }
    }, true, ["sign", "verify"]).then((rsaKey) => {
      return crypto.subtle.exportKey("jwk", rsaKey.privateKey).then(function(rsaPrivateKey) {
        return crypto.subtle.exportKey("jwk", rsaKey.publicKey).then(function(rsaPublicKey) {
            return crypto.subtle.generateKey({
            name: "ECDSA",
            namedCurve: "P-256"
          }, true, ["sign", "verify"]).then((ecdsaKey) => {
            return crypto.subtle.exportKey("jwk", ecdsaKey.privateKey).then(function(ecdsaPrivateKey) {
                return crypto.subtle.exportKey("jwk", ecdsaKey.publicKey).then(function(ecdsaPublicKey) {
              
                // save to extension's localStorage 
                arrSaveToLS("ecdsaPublicKey", {[name]: ecdsaPublicKey});
                arrSaveToLS("ecdsaPrivateKey", {[name]: ecdsaPrivateKey});
                arrSaveToLS("rsaPublicKey", {[name]: rsaPublicKey});
                arrSaveToLS("rsaPrivateKey", {[name]: rsaPrivateKey});
                arrSaveToLS("accountName", name);

                window.localStorage.setItem("activeAccount", name);
                window.localStorage.setItem("provider", that.provider);

                that.ecdsaPublicKey = JSON.stringify(ecdsaPublicKey);
                that.ecdsaKey = ecdsaKey;
                

                return that.setName(name, JSON.stringify(rsaPublicKey), options, callback);
              });
            });
          });
        });
      });
    })
  }

  loadAccounts(callback) {
    let that = this;
    if (callback) {
      return this.loadAccounts_cb(callback);
    } else {
      return new Promise((resolve, reject) => {
        return that.loadAccounts_cb(function(err, result) {
          if (err) {
            reject(err);
          }
          else {
            resolve(result);
            // TODO: also retrun verified in promise
            // resolve([result, verified]);
          }
        });
      });
    }
  }

  loadAccounts_cb(callback) {
    var that = this;
    
    if (!window.localStorage.getItem("activeAccount")) {
      console.log("no account found");
      return callback("no account found", null);
    }
    else {
      var accountName = window.localStorage.getItem("activeAccount");

      this.ecdsaPrivateKey = findKeyInLS(accountName, "ecdsaPrivateKey");
      this.ecdsaPublicKey = findKeyInLS(accountName, "ecdsaPublicKey");
      this.rsaPublicKey = findKeyInLS(accountName, "rsaPublicKey");
      this.rsaPrivateKey = findKeyInLS(accountName, "rsaPrivateKey");
      // console.log(this.ecdsaPrivateKey);
      // console.log(this.ecdsaPublicKey);

      this.ecdsaPublicKey = JSON.stringify(this.ecdsaPublicKey);

      crypto.subtle.importKey("jwk", that.ecdsaPrivateKey, {
          name: "ECDSA",
          namedCurve: "P-256"
        }, false, ['sign'])
        .then(function(result) {
          that.ecdsaKey = {}
          that.ecdsaKey.privateKey = result;
          return callback(null, "load account success");
        })
        .catch((error) => {
          console.log("Error Getting ECDSA Private key: " + error);
          return callback(error.message, null);
        });
    }
  }

  setName(name, rsaPublicKey, options, callback) {
    return this.post(name, {
      "name": name,
      "rsaPublicKey": btoa(rsaPublicKey)
    }, options, callback);
  }

  // Blockchain operations
  getBlock(blockNumber, callback) {
    return this.get(this.joinPath('wolk', 'block', blockNumber.toString(10)), {}, callback);
  }
  getNode(nodeNumber, callback) {
    return this.get(this.joinPath('wolk', 'node', nodeNumber.toString(10)), {}, callback);
  }
  updateNode(nodeNumber, nodeInfo, callback) {
    return this.patch(this.joinPath('wolk', 'node', nodeNumber.toString(10)), nodeInfo, {}, callback);
  }
  getName(name, callback) {
    return this.get(this.joinPath('wolk', 'name', name), {}, callback);
  }
  getAccount(name, callback) {
    return this.get(this.joinPath('wolk', 'account', name), {}, callback);
  }
  getLatestBlockNumber(callback) {
    return this.get(this.joinPath('wolk', 'block', 'latest'), {}, callback);
  }
  getTransaction(txhash, callback) {
    return this.get(this.joinPath('wolk', 'tx', txhash), {}, callback);
  }
  transfer(recipient, amount, callback) {
    return this.post(this.joinPath('wolk', 'transfer'), {
      "recipient": recipient,
      "amount": amount
    }, {}, callback);
  }
  // NoSQL
  createBucket(owner, collection, callback) {
    return this.post(this.joinPath(owner, collection), JSON.stringify({
      "name": collection,
      "bucketType": BucketCloudstore,
      "requesterPays": 0,
      "encryption": "none"
    }), {}, callback);
  }
    
  createCollection(owner, collection, hdrs, callback) {
    return this.post(this.joinPath(owner, collection), JSON.stringify({
      "name": collection,
      "bucketType": BucketNoSQL,
      "requesterPays": 0,
      "encryption": "none"
    }), hdrs, callback);
  }
  listCollections(owner, hdrs, callback) {
    hdrs.Proof = this.determineProofValue();
    return this.get(owner, hdrs, callback);
  }
  deleteCollection(owner, collection, callback) {
    return this.delete(this.joinPath(owner, collection), {}, callback);
  }
  scanCollection(owner, collection, hdrs, callback) {
    hdrs.Proof = this.determineProofValue();
    return this.get(this.joinPath(owner, collection), hdrs, callback);
  }
  setKey(owner, collection, key, value, hdrs, callback) {
    return this.put(this.joinPath(owner, collection, key), value, hdrs, callback);
  }
  getKey(owner, collection, key, hdrs, callback) {
    hdrs.Proof = this.determineProofValue();
    return this.get(this.joinPath(owner, collection, key), hdrs, callback);
  }
  deleteKey(owner, collection, key, callback) {
    return this.delete(this.joinPath(owner, collection, key), {}, callback);
  }

  // SQL
  createDatabase(owner, database, options, callback) {
    let req = {
      "requesttype": "CreateDatabase",
      "owner": owner,
      "database": database
    }
    return this.post(this.joinPath(owner, database, "SQL"), req, options, callback);
  }

  listDatabases(owner, hdrs, callback) {
    let req = {
      "requesttype": "ListDatabases",
      "owner": owner,
      "database": "dummy"
    }
    return this.patch(this.joinPath(owner, "dummy", "SQL"), req, hdrs, callback);
  }

  deleteDatabase(owner, database, callback) {
    let req = {
      "requesttype": "DropDatabase",
      "owner": owner,
      "database": database
    }
    return this.post(this.joinPath(owner, database, "SQL"), req, {}, callback);
  }

  executeSQL(owner, database, sql, hdrs, callback) {
    let req = {
      "requesttype": "Query",
      "owner": owner,
      "database": database,
      "query": sql
    }
    return this.post(this.joinPath(owner, database, "SQL"), req, hdrs, callback);
  }

  readSQL(owner, database, sql, hdrs, callback) {
    let req = {
      "requesttype": "Query",
      "owner": owner,
      "database": database,
      "query": sql
    }
    return this.patch(this.joinPath(owner, database, "SQL"), req, hdrs, callback);
  }
  createTable(owner, database, table, columns, hdrs, callback) {
    let req = {
      "requesttype": "CreateTable",
      "owner": owner,
      "database": database,
      "table": table,
      "columns": columns
    }
    return this.post(this.joinPath(owner, database, "SQL"), req, hdrs, callback);
  }
  listTables(owner, database, hdrs, callback) {
    let req = {
      "requesttype": "ListTables",
      "owner": owner,
      "database": database
    }
    return this.patch(this.joinPath(owner, database, "SQL"), req, hdrs, callback);
  }
  describeTable(owner, database, table, hdrs, callback) {
    let req = {
      "requesttype": "DescribeTable",
      "owner": owner,
      "database": database,
      "table": table
    }
    return this.patch(this.joinPath(owner, database, "SQL"), req, hdrs, callback);
  }
  dropTable(owner, database, table, hdrs, callback) {
    let req = {
      "requesttype": "DropTable",
      "owner": owner,
      "database": database,
      "table": table
    }
    return this.post(this.joinPath(owner, database, "SQL"), req, hdrs, callback);
  }

  // NOT EXPOSED
  get(path, hdrs, callback) {
    return this.wrequest("GET", path, null, hdrs, callback);
  }

  post(path, body, hdrs, callback) {
    return this.wrequest("POST", path, body, hdrs, callback);
  }

  put(path, body, hdrs, callback) {
    return this.wrequest("PUT", path, body, hdrs, callback);
  }

  patch(path, body, hdrs, callback) {
    return this.wrequest("PATCH", path, body, hdrs, callback);
  }

  delete(path, hdrs, callback) {
    return this.wrequest("DELETE", path, null, hdrs, callback);
  }

  sleep(ms) {
    return new Promise(resolve => setTimeout(resolve, ms));
  }

  async waitForTx(txhash, callback) {
    let done = false;
    let calls = 0;
    while (!done) {
      this.getTransaction(txhash, function(err, body) {
        if (err) {
          //console.log("err:", err);
        } else {
          try {
            let tx = JSON.parse(body)
            // if (tx.BlockNumber != undefined) {
            if (tx.BlockNumber) {
              if (! done) {
                done = true;
                return callback(null, tx);
              } else {
                // console.log("repeat");
              }
            }
          } catch (e) {

          }
        }
      })
      await this.sleep(1000);
      if (calls++ > 250) {
        done = true;
        return callback(new Error('timeout, tx not included'), null);
      }
    }
  }

  wrequest(method, path, body, hdrs, callback) {
    let that = this;
    if (callback) {
      return this.wreq(method, path, body, hdrs, callback);
    } else {
      return new Promise((resolve, reject) => {
        return that.wreq(method, path, body, hdrs, function(err, result) {
          if (err) {
            reject(err);
          }
          else {
            resolve(result);
          }
        });
      });
    }
  }

  wreq(method, path, body, hdrs, callback) {
    let enc = new TextEncoder("utf-8")
    let payloadBytes = method + "/" + path;
    if ((method == "POST") || (method == "PATCH")) {
      if (typeof body === "object") {
        body = JSON.stringify(body);
      }
      if (body != null) {
        payloadBytes += body;
      }
    }
    let payload = enc.encode(payloadBytes);
    let that = this
    var ProofValue = 0;

    window.crypto.subtle.sign({
        name: "ECDSA",
        hash: {
          name: "SHA-256"
        }
      }, this.ecdsaKey.privateKey, payload)
      .then(function(signature) {
        var requestOptions = {
          method: method,
          headers: {
            "Sig": buf2hex(signature),
            "Requester": that.ecdsaPublicKey,
            "Msg": payloadBytes
          },
          body: body
        }
        for (var hdr in hdrs) {
          if (hdr == "Proof" && hdrs["Proof"] == 1) {
            ProofValue = 1;
          }
          
          requestOptions.headers[hdr] = hdrs[hdr];
        }
        let url = that.provider + "/" + path
        let proof = ''
        let proofType = ''

        fetch(url, requestOptions)
          .then(function(response) {
            if (response.ok) {
              if (response.headers.has('proof') && response.headers.has('proof-type')) {
                proof = response.headers.get('proof')
                proofType = response.headers.get('proof-type')
              }
              return response.text()
            } else {
              console.log("ERROR:", response.status, response.statusText)
              let statusText = response.statusText;
              if (response.status == 404) statusText = "Not Found";
              let errMsg = "ERROR: " + response.status + " " + statusText;
              
              return callback(errMsg, null);
            }
          })
          .then(function(body) {
            if ( ProofValue != 1 ) {
              return callback(null, body);
            } else {
              if( proof == undefined ) {
                return callback("ERROR: proof requested but not returned", body);
              }
              if( proof.length > 0 ) {
                verifyHeaders(path, hdrs, proof, proofType)
                .then(function(verified) {
                    if (verified) {
                      return callback(null, body, proof, true);
                    } else {
                      return callback("ERROR: requested proof verification failed", body, proof, false);
                    }
                })
                .catch(function(err) {
                  callback(err.message, null)
                })
              } else {
                return callback("ERROR: proof requested but not returned", body);
              }
            }
          })
          .catch(function(err) {
            console.error(err.message);
            return callback(err.message, null)
          })
      })
      .catch(function(err) {
        console.error(err.message);
        return callback(err.message, null)
      });
  }

  joinPath(...args) {
	  return args.map((part, i) => {
	    if (i === 0) {
	      return part.trim().replace(/[\/]*$/g, '')
	    } else {
	      return part.trim().replace(/(^[\/]*|[\/]*$)/g, '')
	    }
	  }).filter(x => x.length).join('/')
	}
}

function arrSaveToLS(key, value) {
  // oldValue is an array
  let oldValue = JSON.parse(window.localStorage.getItem(key));
  oldValue.push(value);
  window.localStorage.setItem(key, JSON.stringify(oldValue));
}

function findKeyInLS(accountName, keyName) {
  let arr = JSON.parse(window.localStorage.getItem(keyName));
  for (var i=0; i < arr.length; i++) {
    if (arr[i][accountName]) {
        return arr[i][accountName];
    }
  }
}

function buf2hex(buffer) { // buffer is an ArrayBuffer
  return Array.prototype.map.call(new Uint8Array(buffer), x => ('00' + x.toString(16)).slice(-2)).join('');
}

function isNotFound(err) {
  return( err.includes("404") ||  err.includes("500") || err.includes("Not Found") );
}

function Computehash(b) {
  return crypto.subtle.digest("SHA-256", b)
}

function hexStringToByte(str) {
  if (!str) {
    return new Uint8Array();
  }
  if ((str[0] == "0") && (str[1] == "x")) {
    str = str.substr(2, str.length - 2);
  }
  var a = [];
  for (var i = 0, len = str.length; i < len; i += 2) {
    a.push(parseInt(str.substr(i, 2), 16));
  }
  return new Uint8Array(a);
}

function getBit(k, i) {
  let bitn = i % 8;
  let byten = 20 - 1 - (i >> 3);
  let res = ((1 << bitn) & k[byten]) > 0;
  return res;
}

function equalBytes(buf1, buf2) {
  if (buf1.byteLength != buf2.byteLength) return false;
  var dv1 = new Int8Array(buf1);
  var dv2 = new Int8Array(buf2);
  for (var i = 0; i != buf1.byteLength; i++) {
    if (dv1[i] != dv2[i]) return false;
  }
  return true;
}

function deserializeProof(sp) {
  let proofhashes = [];
  for (var i = 0; i < sp.Proof.length; i++) {
    proofhashes.push(hexStringToByte(sp.Proof[i]));
  }
  let p = {
    "SMTTreeDepth": 160,
    "Key": hexStringToByte(sp.Key),
    "Proof": proofhashes,
    "ProofBits": hexStringToByte(sp.ProofBits)
  }
  return (p);
}

async function keyToAddress(k) {
  return new Uint8Array(await Computehash(unicodeStringToTypedArray(k))).subarray(0, 20);
}

async function collectionHash(owner, collection) {
  let x = await Computehash(Uint8Array.of(...unicodeStringToTypedArray(owner), ...unicodeStringToTypedArray(collection)));
  return new Uint8Array(x.slice(0, 20))
}

function parseWolkUrl(wolkurl) {
  let urlParts = wolkurl.split("/");
  var owner = ""
  var collection = ""
  var key = ""
  var wolkUrlObject = {
    owner: "",
    collection: "",
    path: "",
  };
  /* wolk://owner/collection/key */

  if (urlParts.length > 0) {
      wolkUrlObject.owner = urlParts.shift();
  }
  if (urlParts.length > 0) {
      wolkUrlObject.collection = urlParts.shift();
  }
  if (urlParts.length > 0) {
      wolkUrlObject.path = urlParts.join("/")
  }
  return wolkUrlObject;
}

function unicodeStringToTypedArray(s) {
  var escstr = encodeURIComponent(s);
  var binstr = escstr.replace(/%([0-9A-F]{2})/g, function(match, p1) {
    return String.fromCharCode('0x' + p1);
  });
  var ua = new Uint8Array(binstr.length);
  Array.prototype.forEach.call(binstr, function(ch, i) {
    ua[i] = ch.charCodeAt(0);
  });
  return ua;
}

async function computeDefaultHashes(smtTreeDepth) {
  var defaultHashes = new Array();
  defaultHashes[0] = new Uint8Array(await Computehash(new Uint8Array(0)));
  for (var level = 1; level < smtTreeDepth; level++) {
    defaultHashes[level] = await Computehash(Uint8Array.of(...new Uint8Array(defaultHashes[level - 1]), ...new Uint8Array(defaultHashes[level - 1])));
  }
  return defaultHashes
}

async function checkSMTProof(p, key, val, root, verbose) {
  var defaultHashes = await computeDefaultHashes(160)
  let cur = val;
  let ProofBits = p.ProofBits;
  let pi = 0; // indexes p.Proof
  var combinedHashes;
  for (var i = 0; i < p.SMTTreeDepth; i++) {
    if (getBit(ProofBits, i)) {
      var pb = p.Proof[pi];
      if (getBit(key, i)) {
        cur = await Computehash(Uint8Array.of(...pb, ...new Uint8Array(cur)));
      } else { // i-th bit is "0", so hash with H([]) on the right
        cur = await Computehash(Uint8Array.of(...new Uint8Array(cur), ...pb));
      }
      pi++;
    } else {
      if (getBit(key, i)) { // i-th bit is "1", so hash with H([]) on the left
        cur = await Computehash(Uint8Array.of(...new Uint8Array(defaultHashes[i]), ...new Uint8Array(cur)));
      } else {
        cur = await Computehash(Uint8Array.of(...new Uint8Array(cur), ...new Uint8Array(defaultHashes[i])));
      }
    }
  }
  let merkleRootComputed = buf2hex(cur);
  let merkleRootExpected = buf2hex(root);
  if (merkleRootComputed == merkleRootExpected) {
    if (verbose > 1) console.log("SUCCESS", "merkleRootComputed", merkleRootComputed);
    // TODO: check that this merkleRoot is actually IN the block with GetBlock
    return true;
  }
  console.log("checkSMTProof FAIL-merkleRoot MISMATCH", merkleRootComputed) //  "merkleRootExpected", merkleRootExpected);
  return false;
}

function verifySMTProof(proofObject, verbose) {
  let proof = deserializeProof(proofObject.Proof);
  return checkSMTProof(proof,
    hexStringToByte(proofObject.KeyHash),
    hexStringToByte(proofObject.TxHash),
    hexStringToByte(proofObject.MerkleRoot),
    verbose);
}

async function verifyNoSQLProof(proofType, proofObject, path, contentLen, verbose) {
  let wolkUrlObject = parseWolkUrl(path)
  if (proofObject.Collection != wolkUrlObject.collection) {
    console.log("Collection Mismatch", proofObject.Collection, wolkUrlObject.collection)
    return false;
  }
  let defaultHashes = await computeDefaultHashes(160);
  if (proofType == "NoSQLScan") {
    if ( proofObject.ScanProofs != undefined ) {
      console.log("NoSQLScan with ScanProofs", proofObject.ScanProofs.length)
      for (var i = 0; i < proofObject.ScanProofs.length; i++) {
        let p = proofObject.ScanProofs[i];
        let prf = deserializeProof(p.KeyProof);
        if (proofObject.Collection.length == 0) { // listCollections
          let smtCheck = await checkSMTProof(prf, prf.Key, hexStringToByte(p.TxHash), hexStringToByte(proofObject.SystemMerkleRoot), defaultHashes, verbose);
          if ( ! smtCheck ) {
            return false;
          }
        } else { // scancollection
          let smtCheck = await checkSMTProof(prf, prf.Key, hexStringToByte(p.TxHash), hexStringToByte(proofObject.CollectionMerkleRoot), defaultHashes, verbose);
          if ( ! smtCheck ) {
            return false;
          }
        }
      }
    } else {
      if (proofObject.Collection.length == 0) { // listCollections
        if ( ! ( proofObject.SystemChunkHash === "0x44a00e7ae0499c6e377b95709f3851843976ea6c2a86eff46adfe15608d22005") ) {
          return false;
        }
        let prf = deserializeProof(proofObject.SystemProof)
        let smtCheck = await checkSMTProof(prf, prf.Key, hexStringToByte(proofObject.SystemChunkHash), hexStringToByte(proofObject.KeyMerkleRoot), defaultHashes, verbose);
        if ( ! smtCheck ) {
          return false;
        }
      } else { // scancollection
        if ( ! ( proofObject.CollectionChunkHash === "44a00e7ae0499c6e377b95709f3851843976ea6c2a86eff46adfe15608d22005" ) ) {
          return false;
        }
      }
    }
    console.log("NoSQLScan Proof Verified")
    return true;
  }

  if (proofObject.Key != wolkUrlObject.path) {
    console.log("Key Mismatch", proofObject.Key, "key", wolkUrlObject.path)
    return false;
  }

  // Verify Proof Process: at proof.Key position (which is keyToAddress(Key) where key is supplied by user)
  //   there is proof.TxHash, which hashes up to proof.CollectionMerkleRoot
  let keyProof = deserializeProof(proofObject.KeyProof);
  let keyHash = await keyToAddress(wolkUrlObject.path);
  if (!equalBytes(keyHash, keyProof.Key)) {
    console.log("Key Mismatch", keyProof.Key, "keyHash", keyHash, "path", wolkUrlObject.path);
    return false;
  }

  // position: keyhash, txhash = value hashes up to merkleroot
  let proof_txHash = hexStringToByte(proofObject.TxHash);
  let proof_collectionMerkleRoot = hexStringToByte(proofObject.CollectionMerkleRoot);
  let keyProofCheck = await checkSMTProof(keyProof, keyHash, proof_txHash, proof_collectionMerkleRoot, defaultHashes, verbose)
  if (!keyProofCheck) {
    console.log("KeyProof Verification FAILED");
    return false;
  }
  console.log("KeyProofCheck: ", keyProofCheck);
  if (proofObject.Tx == undefined) {
    console.log("Tx not supplied")
    return false;
  }

  // check that txhash matches provided tx ... TODO: check that pr.TxHash matches the Tx returned ***in RLP encoded form***
  let txhash = hexStringToByte(proofObject.Tx.txhash);
  let owner = wolkUrlObject.owner;
  let txsigner = hexStringToByte(proofObject.Tx.signer);
  /*
    if (!equalBytes(txhash, proof_txHash)) {
      console.log("Tx mismatch", txhash, proof_txHash);
      return false;
    }
  let txAmount = proofObject.Tx.amount;
  let contentLength = contentLen;
  if (contentLength != txAmount) {
    console.log("Header:Content-Length Does not Match Tx.Amount", "Header:Content-Length", contentLength, "Tx.Amount", txAmount);
    return false;
  }

  // TODO: check that TxSigner is valid, matches owner
  if (!equalBytes(txsigner, owner)) {
    // TODO: check against writers list?
    console.log("Signer mismatch");
    return false
  }
  if (!verifyContentHash(proofObject.Tx.GenericHash, body)) {
    console.log("ContentHash Mismatch");
    return false;
  }
  */

  // Verify Proof Process: at proof.Key position (which is proof.CollectionHash=CollectionHash(owner, collection))
  //   there is proof.CollectionChunkHash, which hashes up to proof.KeyMerkleRoot
  let collectionProof = deserializeProof(proofObject.CollectionProof);
  let collHash = await collectionHash(owner, wolkUrlObject.collection)
  if (!equalBytes(collectionProof.Key, collHash)) {
    console.log("Collection Mismatch", collectionProof.Key, "collHash", collHash);
    return false;
  }

  let proof_collectionChunkHash = hexStringToByte(proofObject.CollectionChunkHash);
  let proof_keyMerkleRoot = hexStringToByte(proofObject.KeyMerkleRoot);
  let collectionCheck = await checkSMTProof(collectionProof, collHash, proof_collectionChunkHash, proof_keyMerkleRoot, defaultHashes, verbose)
  if ( ! collectionCheck ) {
    console.log("CollectionProof Verification FAILED", collectionProof);
    return false;
  }
  console.log("CollectionProof Verified");

  // Verify proof process: at proof.Key position (which is proof.SystemHash=CollectionHash(owner, SystemCollection))
  //   there is the value proof.SystemChunkHash, which hashes up to proof.KeyMerkleRoot
  let systemProof = deserializeProof(proofObject.SystemProof)
  let systemHash = await keyToAddress(owner)
  if (!equalBytes(systemProof.Key, systemHash)) {
    console.log("System Mismatch")
    return false;
  }
  let proof_systemChunkHash = hexStringToByte(proofObject.SystemChunkHash);
  //let proof_keyMerkleRoot = hexStringToByte(proofObject.KeyMerkleRoot);
  let ownerCheck = await checkSMTProof(systemProof, systemHash, proof_systemChunkHash, proof_keyMerkleRoot, defaultHashes, verbose)
  if ( ! ownerCheck ) {
    console.log("SystemProof FAILED")
    return false;
  }
  console.log("SystemProof Verified");
  return true;
}

function verifyHeaders(path, requestHeaders, proof, proofType) {
  if (requestHeaders.Proof != undefined) {
    switch (proofType) {
      case "SMT":
        let smtProof = JSON.parse(proof)
        return verifySMTProof(smtProof, 0);
      case "NoSQLScan":
      case "NoSQL":
        let nosqlProof = JSON.parse(proof)
        return verifyNoSQLProof(proofType, nosqlProof, path, 0, 0);
    }
  } else {
    // console.log("requested proof, but no proof in response");
    return Promise.resolve(true);
  }
  return Promise.resolve(false);
}

if (typeof window !== 'undefined' && typeof window.WEXTENSION === 'undefined') {
    window.WEXTENSION = WEXTENSION;
}

if (typeof window !== 'undefined' && typeof window.verifySMTProof === 'undefined') {
    window.verifySMTProof = verifySMTProof;
}

if (typeof window !== 'undefined' && typeof window.verifyNoSQLProof === 'undefined') {
    window.verifyNoSQLProof = verifyNoSQLProof;
}

// module.exports = WEXTENSION;

module.exports = {
    WEXTENSION: WEXTENSION,
    verifySMTProof: verifySMTProof,
    verifyNoSQLProof: verifyNoSQLProof
}
