var assert = chai.assert;

function randomString(n) {
  return Math.random().toString(36).replace(/[^a-z]+/g, '').substr(0, 5);
}

let TEST_MAX_MS = 20000;
let OWNER1 = "account" + randomString(5);
let OWNER2 = "account" + randomString(5);
let COLLECTION1 = "collection" + randomString(5);
let COLLECTION2 = "collection" + randomString(5);


describe(OWNER1 + ' Account Creation', function() {
  this.timeout(TEST_MAX_MS);

  // option 2: getName and createAccount 2 separate tests
  it('getName(' + OWNER1 + ') not found', function(done) {
    wolkjs.getName(OWNER1, function(err, address) {
      assert(isNotFound(err))
      done();
    });
  });

  it('createAccount(' + OWNER1 + ')', function(done) {
    console.log("wolk.test => creating " + OWNER1)
    wolkjs.createAccount(OWNER1, function(err, txhash) {
      console.log("txhash for createAccount " + OWNER1 + " = " + txhash)
      wolkjs.waitForTx(txhash, function(err, tx) {
        if (err) {
          throw err;
        }
        console.log("===> setName TX: ", txhash, tx)
        done();
      })
    });
  });

  it('getName(' + OWNER1 + ')', function(done) {
    wolkjs.getName(OWNER1, function(err, address) {
      if (err) {
        throw err;
      }
      console.log("getName " + OWNER1 + " : ", address)
      assert(address.length == 40)
      done();
    });
  });
});


describe(OWNER1 + ' Creates Own Collection ', function() {
  this.timeout(TEST_MAX_MS);
  it('listCollections(' + OWNER1 + ') [collections.length=0]', function(done) {
    wolkjs.listCollections(OWNER1, {}, function(err, result) {
      let collections = JSON.parse(result);
      console.log("collections:", collections)
      assert(collections.length == 0)
      done();
    })
  });

  it('createCollection(' + OWNER1 + ', ' + COLLECTION1 + ')', function(done) {
    wolkjs.createCollection(OWNER1, COLLECTION1, function(err, txhash) {
      assert(txhash.length >= 64);
      wolkjs.waitForTx(txhash, function(err, tx) {
        if (err) {
          throw err;
        }
        console.log("===> createCollection TX:", txhash, tx)
        done();
      });
    })
  });

  it('listCollections(' + OWNER1 + ') [collections.length=1]', function(done) {
    wolkjs.listCollections(OWNER1, {}, function(err, result) {
      console.log("result:", result)
      let collections = JSON.parse(result);
      console.log("collections:", collections)
      assert(collections.length == 1)
      done();
    })
  });
})


describe(OWNER2 + ' Account Creation', function() {
  this.timeout(TEST_MAX_MS);

  // option 2: getName and createAccount 2 separate tests
  it('getName(' + OWNER2 + ') not found', function(done) {
    wolkjs.getName(OWNER2, function(err, address) {
      assert(isNotFound(err))
      done();
    });
  });

  it('createAccount(' + OWNER2 + ')', function(done) {
    console.log("wolk.test => creating " + OWNER2)
    wolkjs.createAccount(OWNER2, function(err, txhash) {
      console.log("txhash for createAccount " + OWNER2 + " = " + txhash)
      wolkjs.waitForTx(txhash, function(err, tx) {
        if (err) {
          throw err;
        }
        console.log("===> setName TX: ", txhash, tx)
        done();
      })
    });
  });

  it('getName(' + OWNER2 + ')', function(done) {
    wolkjs.getName(OWNER2, function(err, address) {
      if (err) {
        throw err;
      }
      console.log("getName " + OWNER2 + " : ", address)
      assert(address.length == 40)
      done();
    });
  });
});


describe(OWNER2 + ' Creates Collection for ' + OWNER1, function() {
  this.timeout(TEST_MAX_MS);
  it('listCollections(' + OWNER1 + ') [collections.length=1] with ' + OWNER2 + ' keys', function(done) {
    wolkjs.listCollections(OWNER1, {}, function(err, result) {
      console.log("txhash:", result)
      assert(err == 'ERROR: Permission Denied'); // or other err message
      done();
    })
  });

  it('createCollection(' + OWNER1 + ', ' + COLLECTION2 + ') with ' + OWNER2 + ' keys', function(done) {
    wolkjs.createCollection(OWNER1, COLLECTION2, function(err, txhash) {
      console.log("result:", txhash)
      assert(err == 'ERROR: Permission Denied'); // or other err message
      done();
    })
  });

  it('listCollections(' + OWNER1 + ') [collections.length=1] with ' + OWNER2 + ' keys', function(done) {
    wolkjs.listCollections(OWNER1, {}, function(err, result) {
      console.log("result:", result)
      assert(err == 'ERROR: Permission Denied'); // or other err message
      done();
    })
  });
})
