var assert = chai.assert;

function randomString(n) {
  return Math.random().toString(36).replace(/[^a-z]+/g, '').substr(0, 5);
}

let TEST_MAX_MS = 20000;
let OWNER = "account" + randomString(5);
let COLLECTION = "collection" + randomString(5);

// getLatestBlockNumber(callback)
describe('Get Blocks', function() {
  this.timeout(TEST_MAX_MS);
  let BLKNUM = 1;
  it('getLatestBlockNumber()', function(done) {
    wolkjs.getLatestBlockNumber(function(err, result, verified) {
      assert(!isNaN(result));
      console.log("getLatestBlockNumber:", result)
      BLKNUM = result
      done();
    });
  });
  it('getBlock()', function(done) {
    wolkjs.getBlock(BLKNUM, function(err, result, verified) {
      let block = JSON.parse(result);
      assert(block.blockNumber == BLKNUM);
      console.log("Block # " + block.blockNumber + ":" + result)
      done();
    });
  });
});

describe('Account Creation', function() {
  this.timeout(TEST_MAX_MS);
  it('getName(' + OWNER + ') not found', function(done) {
    wolkjs.getName(OWNER, function(err, address, verified) {
      assert(isNotFound(err))
      done();
    });
  });

  it('createAccount(' + OWNER + ')', function(done) {
    console.log("wolk.test => creating " + OWNER)
    wolkjs.createAccount(OWNER, function(err, txhash) {
      console.log("txhash for createAccount = " + txhash)
      wolkjs.waitForTx(txhash, function(err, tx) {
        if (err) {
          throw err;
        }
        console.log("===> setName TX: ", txhash, tx)
        done();
      })
    });
  });

  it('getName(' + OWNER + ')', function(done) {
    wolkjs.getName(OWNER, function(err, address) {
      if (err) {
        throw err;
      }
      console.log("getName: ", address) // TODO check against address derived from ECDSA public key
      assert(address.length == 40)
      done();
    });
  });
});


// transfer(recipient, amount)
describe('Balance Transfer', function() {
  this.timeout(TEST_MAX_MS*2);
  let SENDER = "account" + randomString(5);
  let AMOUNT = 5000;
  let STARTING_BALANCE = 25000;
  let EXPECTED_SENDER_BALANCE = STARTING_BALANCE - AMOUNT;
  let EXPECTED_OWNER_BALANCE = STARTING_BALANCE + AMOUNT;
  it('transfer(' + OWNER + ', ' + AMOUNT + ')', function(done) {
    wolkjs.createAccount(SENDER, function(err, txhash) {
      wolkjs.waitForTx(txhash, function(err, tx) {
        if (err) {
          throw err;
        }
        console.log("===> createAccount(SENDER) TX: ", txhash, tx)
        wolkjs.transfer(OWNER, AMOUNT, function(err, txhash) {
          wolkjs.waitForTx(txhash, function(err, tx) {
            if (err) {
              throw err;
            }
            console.log("===> Transfer TX: ", txhash, tx)
            done();
          })
        });
      });
    });
  });

  it('getAccount(' + OWNER + ')', function(done) {
    wolkjs.getAccount(OWNER, function(err, ownerJSON) {
      var account_owner = JSON.parse(ownerJSON);
      console.log("ACCOUNT_OWNER", account_owner)
      console.log(account_owner.balance, EXPECTED_SENDER_BALANCE);
      assert(account_owner.balance === EXPECTED_OWNER_BALANCE);
      done();
    });
  })


  it('getAccount(' + SENDER + ')', function(done) {
    wolkjs.getAccount(SENDER, function(err, senderJSON) {
      var account_sender = JSON.parse(senderJSON);
      console.log("ACCOUNT_SENDER", account_sender)
      console.log(account_sender.balance, EXPECTED_SENDER_BALANCE);
      assert(account_sender.balance === EXPECTED_SENDER_BALANCE);
      done();
    });
  })
});

describe('Collection Creation', function() {
  this.timeout(TEST_MAX_MS);
  it('listCollections(' + OWNER + ') [collections.length=0]', function(done) {
    wolkjs.listCollections(OWNER, {}, function(err, result) {
      let collections = JSON.parse(result);
      console.log("collections:", collections)
      assert(collections.length == 0)
      done();
    })
  });

  it('createCollection(' + OWNER + ', ' + COLLECTION + ')', function(done) {
    wolkjs.createCollection(OWNER, COLLECTION, function(err, txhash) {
      assert(txhash.length >= 64);
      wolkjs.waitForTx(txhash, function(err, tx) {
        if (err) {
          throw err;
        }
        console.log("===> createCollection TX:", txhash, tx)
        done();
      });
    })
  });

  it('listCollections(' + OWNER + ') [collections.length=1]', function(done) {
    wolkjs.listCollections(OWNER, {}, function(err, result) {
      console.log("result:", result)
      let collections = JSON.parse(result);
      console.log("collections:", collections)
      assert(collections.length == 1)
      done();
    })
  });
})

describe('Key Addition/Deletion', function() {
  this.timeout(TEST_MAX_MS);
  let KEY = "key" + randomString(5);
  let VAL = "val" + randomString(5);

  it('getKey(' + OWNER + ', ' + COLLECTION + ', ' + KEY + ') not found', function(done) {
    wolkjs.getKey(OWNER, COLLECTION, KEY, {}, function(err, result) {
      assert(isNotFound(err));
      done();
    })
  })

  it('setKey(' + OWNER + ', ' + COLLECTION + ', ' + KEY + ', ' + VAL + ')', function(done) {
    wolkjs.setKey(OWNER, COLLECTION, KEY, VAL, function(err, txhash) {
      assert(txhash.length >= 64);
      wolkjs.waitForTx(txhash, function(err, tx) {
        if (err) {
          throw err;
        }
        console.log("===> setKey TX:", txhash, tx)
        done();
      })
    })
  });

  it('getKey(' + OWNER + ', ' + COLLECTION + ', ' + KEY + ') match VAL', function(done) {
    wolkjs.getKey(OWNER, COLLECTION, KEY, {}, function(err, result) {
      console.log("getKey: ", result);
      assert(result == VAL);
      done();
    })
  })

  it('scanCollection(' + OWNER + ', ' + COLLECTION + ') [elements.length=1]', function(done) {
    wolkjs.scanCollection(OWNER, COLLECTION, {}, function(err, result) {
      console.log("result:", result);
      let elements = JSON.parse(result)
      assert(elements.length == 1)
      done();
    });
  })

  it('deleteKey(' + OWNER + ', ' + COLLECTION + ', ' + KEY + ')', function(done) {
    wolkjs.deleteKey(OWNER, COLLECTION, KEY, function(err, txhash) {
      console.log("txhash:", txhash)
      assert(txhash.length >= 64);
      wolkjs.waitForTx(txhash, function(err, tx) {
        if (err) {
          throw err;
        }
        console.log("===> deleteKey TX:", txhash, tx)
        done();
      })
    });
  });

  it('scanCollection(' + OWNER + ', ' + COLLECTION + ')  [elements.length=0]', function(done) {
    wolkjs.scanCollection(OWNER, COLLECTION, {}, function(err, result) {
      console.log("result:", result);
      let elements = JSON.parse(result)
      assert(elements.length == 0)
      done();
    });
  })
});

describe('Collection Deletion', function() {
  this.timeout(TEST_MAX_MS);
  it('deleteCollection(' + OWNER + ", " + COLLECTION + ')', function(done) {
    wolkjs.deleteCollection(OWNER, COLLECTION, function(err, txhash) {
      assert(txhash.length >= 64);
      wolkjs.waitForTx(txhash, function(err, tx) {
        if (err) {
          throw err;
        }
        console.log("===> deleteCollection TX:", txhash, tx)
        done();
      })
    })
  });
  it('listCollections(' + OWNER + ') [collections.length=1]', function(done) {
    wolkjs.listCollections(OWNER, {}, function(err, result) {
      let collections = JSON.parse(result);
      console.log("collections:", result)
      assert("length", collections.length == 1)
      done();
    });
  });
});

describe('Node Info', function() {
  this.timeout(TEST_MAX_MS);
  let NODE_NUMBER = 1;
  it('getNode', function (done) {
      wolkjs.getNode(NODE_NUMBER, function (err, result, txhash) {
        if (err) {
          throw err;
        }
        // {"owner":"0x8520f2be32a14fd8eba2773a23ecd1ebd2290b99","valueInt":10000,"storageip":"35.245.174.47","consensusip":"0.0.0.0","region":1,"httpport":0}
        let node = JSON.parse(result)
        assert(node.valueInt == 10000)
        done();
      });
  });
})
/*
describe('Node Info', function() {
  it('updateNode', function (done) {
      let nodeInfo = {"storageip": "35.245.174.47", "consensusip": "35.245.174.47", "region": 4, "valueInt": 123456};
      wolkjs.updateNode(NODE_NUMBER, nodeInfo, function (err, txhash) {
        wolkjs.waitForTx(txhash, function (err, tx) {
          if ( err ) { throw err; }
          console.log("===> updateNode TX: ", txhash, tx)
          /*wolkjs.getNode(NODE, function (err, result, verified) {
            console.log(result);
            var resultJson = JSON.parse(result);
            assert(resultJson.owner === "0x8520f2be32a14fd8eba2773a23ecd1ebd2290b99");
            assert(resultJson.valueInt === 10000);
            assert(resultJson.storageip === STORAGEIP);
            assert(resultJson.consensusip === CONSENSUSIP);
            assert(resultJson.httpport === 0);
            assert(resultJson.region === 1);
            done();
        });
      });
  });
});
*/
