var assert = chai.assert;

function randomString(n) {
  return Math.random().toString(36).replace(/[^a-z]+/g, '').substr(0, 5);
}

let TEST_MAX_MS = 20000;
let OWNER = "account" + randomString(5);
let COLLECTION = "collection" + randomString(5);

// getLatestBlockNumber(callback)
describe('Get Blocks', function() {
  this.timeout(TEST_MAX_MS);
  let BLKNUM = 1;
  it('getLatestBlockNumber()', function(done) {
    wolkjs.getLatestBlockNumber()
      .then((result) => {
        assert(!isNaN(result));
        console.log("getLatestBlockNumber:", result)
        BLKNUM = result
        done();
      });
  });
  it('getBlock()', function(done) {
    wolkjs.getBlock(BLKNUM)
      .then((result) => {
        let block = JSON.parse(result);
        assert(block.blockNumber == BLKNUM);
        console.log("Block # " + block.blockNumber + ":" + result)
        done();
      });
  });
});

describe('Account Creation', function() {
  this.timeout(TEST_MAX_MS);

  it('getName(' + OWNER + ') not found', function(done) {
    wolkjs.getName(OWNER)
      .catch(function(err) {
        assert(isNotFound(err))
        done();
      });
  });

  it('createAccount(' + OWNER + ')', function(done) {
    wolkjs.createAccount(OWNER, function(err, txhash) {
      wolkjs.waitForTx(txhash, function(err, tx) {
        if (err) {
          throw err;
        }
        console.log("===> setName TX: ", txhash, tx)
        done();
      })
    });
  });

  it('getName(' + OWNER + ') found', function(done) {
    wolkjs.getName(OWNER)
      .then(function(result) {
        console.log(result)
        done();
      });
  });

});

describe('Collection Creation', function() {
  this.timeout(TEST_MAX_MS);
  it('listCollections(' + OWNER + ') [collections.length=1]', function(done) {
    wolkjs.listCollections(OWNER, {})
      .then((result) => {
        let collections = JSON.parse(result);
        console.log("collections:", collections)
        assert(collections.length == 0)
        // assert(verified);
        done();
      })
  });

  it('createCollection(' + OWNER + ', ' + COLLECTION + ')', function(done) {
    wolkjs.createCollection(OWNER, COLLECTION)
      .then((txhash) => {
        assert(txhash.length >= 64);
        wolkjs.waitForTx(txhash, function(err, tx) {
          if (err) {
            throw err;
          }
          console.log("===> createCollection TX:", txhash, tx)
          done();
        });
      })
  });

  it('listCollections(' + OWNER + ') [collections.length=2]', function(done) {
    wolkjs.listCollections(OWNER, {})
      .then((result) => {
        let collections = JSON.parse(result);
        console.log("collections:", collections)
        assert(collections.length == 1)
        // assert(verified);
        done();
      })
  });
})

describe('Key Addition/Deletion', function() {
  this.timeout(TEST_MAX_MS);
  let KEY = "key" + randomString(5);
  let VAL = "val" + randomString(5);


  it('getKey(' + OWNER + ', ' + COLLECTION + ', ' + KEY + ')', function(done) {
    wolkjs.getKey(OWNER, COLLECTION, KEY, {})
      .catch(function(err) {
        assert(isNotFound(err))
        done();
      });
  })

  it('setKey(' + OWNER + ', ' + COLLECTION + ', ' + KEY + ', ' + VAL + ')', function(done) {
    wolkjs.setKey(OWNER, COLLECTION, KEY, VAL)
      .then(txhash => {
        assert(txhash.length >= 64);
        wolkjs.waitForTx(txhash, function(err, tx) {
          if (err) {
            throw err;
          }
          console.log("===> setKey TX:", txhash, tx)
          done();
        })
      })
  });

  it('getKey(' + OWNER + ', ' + COLLECTION + ', ' + KEY + ')', function(done) {
    wolkjs.getKey(OWNER, COLLECTION, KEY, {})
      .then(result => {
        console.log("getKey: ", result);
        assert(result == VAL);
        // assert(verified);
        done();
      })
  })

  it('scanCollection(' + OWNER + ', ' + COLLECTION + ') [elements.length=1]', function(done) {
    wolkjs.scanCollection(OWNER, COLLECTION, {})
      .then(result => {
        console.log("result:", result);
        let elements = JSON.parse(result)
        assert(elements.length == 1)
        // assert(verified);
        done();
      });
  })

  it('deleteKey(' + OWNER + ', ' + COLLECTION + ', ' + KEY + ')', function(done) {
    wolkjs.deleteKey(OWNER, COLLECTION, KEY)
      .then(txhash => {
        console.log("txhash:", txhash)
        assert(txhash.length >= 64);
        wolkjs.waitForTx(txhash, function(err, tx) {
          if (err) {
            throw err;
          }
          console.log("===> deleteKey TX:", txhash, tx)
          done();
        })
      });
  });

  it('scanCollection(' + OWNER + ', ' + COLLECTION + ')  [elements.length=0]', function(done) {
    wolkjs.scanCollection(OWNER, COLLECTION, {})
      .then(result => {
        console.log("result:", result);
        let elements = JSON.parse(result)
        assert(elements.length == 0)
        // assert(verified);
        done();
      });
  })
});

describe('Collection Deletion', function() {
  this.timeout(TEST_MAX_MS);
  it('deleteCollection(' + OWNER + ", " + COLLECTION + ')', function(done) {
    wolkjs.deleteCollection(OWNER, COLLECTION)
      .then(txhash => {
        assert(txhash.length >= 64);
        wolkjs.waitForTx(txhash, function(err, tx) {
          if (err) {
            throw err;
          }
          console.log("===> deleteCollection TX:", txhash, tx)
          done();
        })
      })
  });
  it('listCollections(' + OWNER + ') [collections.length=1]', function(done) {
    wolkjs.listCollections(OWNER, {})
      .then(result => {
        let collections = JSON.parse(result);
        console.log("collections:", result)
        assert("length", collections.length == 1)
        done();
      });
  });
});

// transfer(recipient, amount)
describe('Balance Transfer', function() {
  this.timeout(TEST_MAX_MS);
  let SENDER = "account" + randomString(5);
  let AMOUNT = 5000;
  let STARTING_BALANCE = 25000;
  let EXPECTED_SENDER_BALANCE = STARTING_BALANCE - AMOUNT;
  let EXPECTED_OWNER_BALANCE = STARTING_BALANCE + AMOUNT;
  it('transfer(' + OWNER + ', ' + AMOUNT + ')', function(done) {
    wolkjs.createAccount(SENDER, function(err, txhash) {
      wolkjs.waitForTx(txhash, function(err, tx) {
        if (err) {
          throw err;
        }
        console.log("===> createAccount(SENDER) TX: ", txhash, tx)
        wolkjs.transfer(OWNER, AMOUNT)
          .then(txhash => {
            wolkjs.waitForTx(txhash, function(err, tx) {
              if (err) {
                throw err;
              }
              console.log("===> Transfer TX: ", txhash, tx)
              done();
            })
          });
      });
    });
  });

  it('getAccount(' + OWNER + ')', function(done) {
    wolkjs.getAccount(OWNER)
      .then(ownerJSON => {
        var account_owner = JSON.parse(ownerJSON);
        console.log(account_owner)
        assert(account_owner.balance === EXPECTED_OWNER_BALANCE);
        done();
      });
  })


  it('getAccount(' + SENDER + ')', function(done) {
    wolkjs.getAccount(SENDER)
      .then(senderJSON => {
        var account_sender = JSON.parse(senderJSON);
        console.log(account_sender.balance, EXPECTED_SENDER_BALANCE);
        assert(account_sender.balance === EXPECTED_SENDER_BALANCE);
        done();
      });
  })
});


describe('Node Info', function() {
  this.timeout(TEST_MAX_MS);
  let NODE_NUMBER = 1;
  it('getNode', function(done) {
    wolkjs.getNode(NODE_NUMBER)
      .then(result => {
        // {"owner":"0x8520f2be32a14fd8eba2773a23ecd1ebd2290b99","valueInt":10000,"storageip":"35.245.174.47","consensusip":"0.0.0.0","region":1,"httpport":0}
        let node = JSON.parse(result)
        assert(node.valueInt == 10000)
        done();
      });
  });
})

/*
describe('Node Info', function() {
  it('updateNode', function (done) {
      let nodeInfo = {"storageip": "35.245.174.47", "consensusip": "35.245.174.47", "region": 4, "valueInt": 123456};
      wolkjs.updateNode(NODE_NUMBER, nodeInfo, function (err, txhash) {
        wolkjs.waitForTx(txhash, function (err, tx) {
          if ( err ) { throw err; }
          console.log("===> updateNode TX: ", txhash, tx)
          /*wolkjs.getNode(NODE, function (err, result, verified) {
            console.log(result);
            var resultJson = JSON.parse(result);
            assert(resultJson.owner === "0x8520f2be32a14fd8eba2773a23ecd1ebd2290b99");
            assert(resultJson.valueInt === 10000);
            assert(resultJson.storageip === STORAGEIP);
            assert(resultJson.consensusip === CONSENSUSIP);
            assert(resultJson.httpport === 0);
            assert(resultJson.region === 1);
            done();
        });
      });
  });
});
*/
