var assert = chai.assert;

function randomString(n) {
  return Math.random().toString(36).replace(/[^a-z]+/g, '').substr(0, 5);
}

// common tables
let ProfilesCollection = "profiles";   // https://schema.org/Person
let FriendsCollection = "friends";     // https://schema.org/BefriendAction
let PostsCollection = "posts";         // https://schema.org/SocialMediaPosting  https://schema.org/CommentAction
let MailCollection = "mail";           // https://schema.org/EmailMessage
let DeveloperCollection = "webapplication"; // https://schema.org/WebApplication
let AppCollection = "apps";            // https://schema.org/InstallAction

function RSAEncrypt(x) {
  return x
}

// Apps class holds core
class WolkUser extends WOLK {
  constructor(name, callback) {
    super(function(err, result) {
    
    })

    this.createAccount(name, function(err, txhash) {
      if (err) {
        throw err;
      }
      callback(null, txhash);
    })
  }

  getRSAPublicKey(x) {
    return "TODO"
  }

  getFriendsKey() {
    return "TODO"
  }

  // friendship
  befriend(request, callback) {
    return this.put(joinPath(request.object.name, FriendsCollection, request.agent.name), request, callback)
  }

  getFriendRequests(callback) {
    return this.get(joinPath(this.name, FriendsCollection), {"sortOrder":"endTime", "direction": "desc"}, callback)
  }

  acceptFriend(request, callback) {
    return this.patch(joinPath(this.name, FriendsCollection), request, callback)
  }

  rejectFriend(request, callback) {
    return this.delete(joinPath(this.name, FriendsCollection), request, callback)
  }

  getFriends(callback) {
    return this.get(joinPath(this.name, FriendsCollection), {"sortOrder":"lastInteractionTime", "direction": "desc"}, callback)
  }

  // social media posts are publicly visible to friends, but they need the users friendsKey.  X posts (sent by X) are in X's bucket
  postMessage(postAction, callback) {
    return this.post(joinPath(this.name, PostsCollection), postAction, callback)
  }

  commentMessage(commentAction, callback) { // comments/likes are ABOUT a post
    return this.post(joinPath(this.name, PostsCollection, commentAction.messageID, CommentCollection), commentAction, callback)
  }

  getPosts(friend, callback) {
    return this.get(joinPath(friend.name, PostsCollection), {"sortOrder":"dateCreated", "direction": "desc"}, callback)
  }

  // mail messages are visible to receiver only, and the sender needs the RSAPublicKey: X's mail (sent by others) is in X's bucket
  emailMessage(receiver, mailAction, callback) {
    let encryptedMail = RSAEncrypt(receiver.RSAPublicKey, mailAction)
    return this.post(joinPath(this.name, MailCollection, mailAction.messageID), encryptedMail, callback)
  }

  getMailMessages(callback) {
    return this.get(joinPath(friend.name, MailCollection), {"sortOrder":"dateSent", "direction": "desc"}, callback)
  }

  getMailMessage(messageID, callback) {
    return this.get(joinPath(friend.name, MailCollection, mailAction.messageID), {}, callback)
  }

  // developer publishing web applications on the appstore
  publishWebApplication(publishAction) {
    return this.put(joinPath(ProtocolName, DeveloperCollection, publishAction.name), publishAction, callback)
  }

  updateWebApplication(updateAction) {
    return this.patch(joinPath(ProtocolName, DeveloperCollection, publishAction.name), publishAction, callback)
  }

  listWebApplications(callback) {
    return this.get(joinPath(ProtocolName, DeveloperCollection), {"sortOrder":"lastPublishTime"}, callback)
  }

  getWebApplications(name, callback) {
    return this.get(joinPath(ProtocolName, DeveloperCollection, name), {}, callback);
  }

  unpublishWebApplication(name, callback) {
    return this.delete(joinPath(ProtocolName, DeveloperCollection, name), callback);
  }

  // user reviewing web application
  reviewWebApplication(userReview, callback) {  // reviews are about a web application
    return this.put(joinPath(ProtocolName, DeveloperCollection, userReview.name, UserReviewsCollection), userReview, callback);
  }

  // user installing web application
  installApp(installAction, callback) {
    return this.put(joinPath(this.name, AppCollection, installAction.name), installAction, callback)
  }

  listApps(callback) {
    return this.get(joinPath(this.name, AppCollection), {"sortOrder":"installTime", "direction": "desc"}, callback)
  }

  getApp(name, callback) {
    return this.get(joinPath(this.name, AppCollection, name), callback)
  }

  uninstallApp(name, callback) {
    return this.delete(joinPath(this.name, AppCollection, name), callback)
  }

}
